from datetime import datetime, date
from decimal import Decimal

def jsonity(data:dict) -> dict:
    r = {}
    for k, v in data.items():
        if isinstance(v, datetime):
            r[k] = v.isoformat()
        elif isinstance(v, date):
            r[k] = v.isoformat()
        elif isinstance(v, Decimal):
            r[k] = str(v)
        else:
            r[k] = v
    return r


from __future__ import annotations
from typing import List, Dict,Optional
from sqlalchemy.sql.schema import Table as SQLTable
from sqlalchemy.sql.schema import Column as SQLColumn
import sqlalchemy
from devtools import debug
from datetime import date, datetime, timedelta,time
from decimal import Decimal
from enum import Enum
import json
from loguru import logger
from faker import Faker
from faker.providers import python
from databases import Database
from loguru import logger
from devtools import debug
from unimeta.libs.liburl import parse_url
from unimeta.libs.libformat import jsonity
import re

_fake = Faker(['zh_CN'])

def fake(c:Column, hint:dict={}) -> str:
    if c.name in hint:
        f = hint[c.name]
        func = getattr(_fake,f)
        return func()
    elif isinstance(c,StringColumn):
        return _fake.pystr(max_chars=c.length)
    elif isinstance(c, TextColumn):
        return _fake.text()
    elif isinstance(c, IntegerColumn):
        return _fake.pyint()
    elif isinstance(c,DecimalColumn):
        fstr = "%%.%df" % c.scale
        left_digits = c.precision - c.scale 
        return Decimal(fstr %_fake.pyfloat(left_digits=left_digits,positive=True))
    elif isinstance(c, FloatColumn):
        return _fake.pyfloat()
    elif isinstance(c, BooleanColumn):
        return _fake.pybool()
    elif isinstance(c, DateTimeColumn):
        return _fake.date_time()
    elif isinstance(c, DateColumn):
        return _fake.date()
    elif isinstance(c, TimeColumn):
        return _fake.iso8601()
    elif isinstance(c, JSONColumn):
        return  json.dumps(_fake.pydict())
    else:
        return _fake.name()

class CHTableEngine(Enum):
    MergeTree = "MergeTree" 
    ReplacingMergeTree = "ReplacingMergeTree"
    SummingMergeTree = 'SummingMergeTree'
    AggregatingMergeTree = "AggregatingMergeTree"
    CollapsingMergeTree = "CollapsingMergeTree"
    VersionedCollapsingMergeTree = "VersionedCollapsingMergeTree"
    GraphiteMergeTree = "GraphiteMergeTree"
    TinyLog = "TinyLog"
    StripeLog = "StripeLog"
    Log = "Log"
    Kafka = "Kafka"
    MySQL = "MySQL"
    ODBC = "ODBC"
    JDBC = "JDBC"
    HDFS = "HDFS"
    Distributed = "Distributed"
    MaterializedView = "MaterializedView"
    Dictionary = "Dictionary"
    Merge = "Merge"
    File = "File"
    Null = "Null"
    Set = "Set"
    Join = "Join"
    URL = "URL"
    View = "View"
    Memory = "Memory"
    Buffer = "Buffer"



class Column:
    name:str
    nullable:bool
    autoincrement:bool = False
    primary_key:bool = False

    def to_ch_type(self) -> str:
        pass
    
    @classmethod
    def read_from_sqlcolumn(cls,sqlcolumn: SQLColumn) -> Column:
        column = Column()
        column.nullable = sqlcolumn.nullable
        column.name = sqlcolumn.name
        return column

    def avro_types(cls) -> Optional[str,List[str]]:
        pass


class StringColumn(Column):
    length:int
    
    @classmethod
    def read_from_sqlcolumn(cls,sqlcolumn: SQLColumn) -> StringColumn:
        column = StringColumn()
        column.nullable = sqlcolumn.nullable
        column.name = sqlcolumn.name
        column.length = sqlcolumn.type.length
        return column

    def to_ch_type(self) -> str:
        if self.nullable:
            return "Nullable(String)"
        else:
            return "String"

    
    def avro_types(self) -> Optional[str,List[str]]:
        if self.nullable:
            return ["string", "null"]
        else:
            return "string"


class PhoneField(StringColumn):
    pass



class TextColumn(Column):

    def to_ch_type(self) -> str:
        if self.nullable:
            return "Nullable(String)"
        else:
            return "String"

    def avro_types(self) -> Optional[str,List[str]]:
        if self.nullable:
            return ["string", "null"]
        else:
            return "string"


class IntegerColumn(Column):
    autoincrement:bool

    @classmethod
    def read_from_sqlcolumn(cls,sqlcolumn: SQLColumn) -> IntegerColumn:
        column = IntegerColumn()
        column.nullable = sqlcolumn.nullable
        column.name = sqlcolumn.name
        column.autoincrement = sqlcolumn.autoincrement
        return column

    def to_ch_type(self) -> str:
        if self.nullable:
            return "Nullable(UInt64)"
        else:
            return "UInt64"

    def avro_types(self) -> Optional[str,List[str]]:
        if self.nullable:
            return ["int", "null"]
        else:
            return "int"


class DecimalColumn(Column):
    precision:int
    scale:int

    @classmethod
    def read_from_sqlcolumn(cls,sqlcolumn: SQLColumn) -> DecimalColumn:
        column = DecimalColumn()
        column.nullable = sqlcolumn.nullable
        column.name = sqlcolumn.name
        column.precision = sqlcolumn.type.precision
        column.scale = sqlcolumn.type.scale
        return column

    def to_ch_type(self) -> str:
        if self.nullable:
            return "Nullable(String)"
        else:
            return "String"

    def avro_types(self) -> Optional[str,List[str]]:
        if self.nullable:
            return ["bytes", "null"]
        else:
            return "bytes"


class FloatColumn(Column):
    pass

    def to_ch_type(self) -> str:
        if self.nullable:
            return "Nullable(Float32)"
        else:
            return "Float32"

    @classmethod
    def read_from_sqlcolumn(cls,sqlcolumn: SQLColumn) -> FloatColumn:
        column = FloatColumn()
        column.nullable = sqlcolumn.nullable
        column.name = sqlcolumn.name
        return column

    def avro_types(self) -> Optional[str,List[str]]:
        if self.nullable:
            return ["float", "null"]
        else:
            return "float"

class BooleanColumn(Column):
    pass

    def to_ch_type(self) -> str:
        if self.nullable:
            return "Nullable(UInt8)"
        else:
            return "UInt8"

    @classmethod
    def read_from_sqlcolumn(cls,sqlcolumn: SQLColumn) -> BooleanColumn:
        column = BooleanColumn()
        column.nullable = sqlcolumn.nullable
        column.name = sqlcolumn.name
        return column

    def avro_types(self) -> Optional[str,List[str]]:
        if self.nullable:
            return ["boolean", "null"]
        else:
            return "boolean"


class DateTimeColumn(Column):
    pass
    
    def to_ch_type(self) -> str:
        if self.nullable:
            return "Nullable(DateTime)"
        else:
            return "DateTime"

    @classmethod
    def read_from_sqlcolumn(cls,sqlcolumn: SQLColumn) -> DateTimeColumn:
        column = DateTimeColumn()
        column.nullable = sqlcolumn.nullable
        column.name = sqlcolumn.name
        return column

    def avro_types(self) -> Optional[str,List[str]]:
        if self.nullable:
            return ["int", "null"]
        else:
            return "int"



class DateColumn(Column):
    
    @classmethod
    def read_from_sqlcolumn(cls,sqlcolumn: SQLColumn) -> DateColumn:
        column = DateColumn()
        column.nullable = sqlcolumn.nullable
        column.name = sqlcolumn.name
        return column
        
    def to_ch_type(self) -> str:
        if self.nullable:
            return "Nullable(Date)"
        else:
            return "Date"

    
    def avro_types(self) -> Optional[str,List[str]]:
        if self.nullable:
            return ["int", "null"]
        else:
            return "int"



class TimeColumn(Column):
    
    def to_ch_type(self) -> str:
        if self.nullable:
            return "Nullable(String)"
        else:
            return "String"

    def avro_types(self) -> Optional[str,List[str]]:
        if self.nullable:
            return ["int", "null"]
        else:
            return "int"


class JSONColumn(Column):
    
    def to_ch_type(self) -> str:
        if self.nullable:
            return "Nullable(String)"
        else:
            return "String"



    def avro_types(self) -> Optional[str,List[str]]:
        if self.nullable:
            return ["string", "null"]
        else:
            return "string"




def get_column_from_sql(sqlcolumn:SQLColumn) -> Column:
    ptype = sqlcolumn.type.python_type
    if ptype is int:
        return IntegerColumn.read_from_sqlcolumn(sqlcolumn)
    elif ptype is str:
        return StringColumn.read_from_sqlcolumn(sqlcolumn)
    elif ptype is date:
        return DateColumn.read_from_sqlcolumn(sqlcolumn)
    elif ptype is Decimal:
        return DecimalColumn.read_from_sqlcolumn(sqlcolumn)
    elif ptype is datetime:
        return DateTimeColumn.read_from_sqlcolumn(sqlcolumn)
    elif ptype is float:
        return FloatColumn.read_from_sqlcolumn(sqlcolumn)
    elif ptype is time:
        return IntegerColumn.read_from_sqlcolumn(sqlcolumn)
    else:
        debug(ptype)
        debug(sqlcolumn)


def get_column_from_meta(k:str, v:Optional[List[str],str],is_primary:bool) -> Column:
    nullable = False
    if isinstance(v,list):
        v.remove("null")
        nullable = True
        if len(v) > 1:
            raise Exception("error type union")
        v = v[0]
    if v == "string":
        column = StringColumn()
    elif v == "int":
        column = IntegerColumn()
    elif v == "boolean":
        column = BooleanColumn()
    elif v == "float":
        column = FloatColumn()
    elif v == "bytes":
        column =StringColumn()
    else:
        raise Exception("unknown data type")
    column.name = k
    column.nullable = nullable
    column.primary_key = is_primary
    return column
class ClickhouseTableTemplate():

    @classmethod
    def getVersionedCollapsingMergeTree(cls) -> str:
        return """
            CREATE TABLE IF NOT EXISTS {db_name}.{table_name}
            (
                 Sign Int8,
                 Version UInt8,
                {create_column}
            )
            ENGINE = VersionedCollapsingMergeTree(Sign, Version)
            PARTITION BY toYYYYMM({primary_date})
            ORDER BY ({primary_key})
            PRIMARY KEY ({primary_key})
        """

    @classmethod
    def getReplacingMergeTree(cls) -> str:
        return """
            CREATE TABLE IF NOT EXISTS {db_name}.{table_name}
            (
                {create_column}
            )
            ENGINE = ReplacingMergeTree()
            PARTITION BY toYYYYMM({primary_date})
            ORDER BY ({primary_key})
            PRIMARY KEY ({primary_key})
        """

    @classmethod
    def getDefault(cls) -> str:
        return """
            CREATE TABLE IF NOT EXISTS  {db_name}.{table_name}
            (
                {create_column}
            )
            ENGINE = {engine_name}()
            PARTITION BY toYYYYMM({primary_date})
            ORDER BY ({primary_key})
        """

    @classmethod
    def get(cls,engine:CHTableEngine)->str:
        if engine is CHTableEngine.VersionedCollapsingMergeTree:
            return cls.getVersionedCollapsingMergeTree()
        elif engine is CHTableEngine.ReplacingMergeTree:
            return cls.getReplacingMergeTree()
        else:
            return cls.getDefault()


class DDLTemplate():
    @classmethod
    def get_insert_template(cls) -> str:
        return """
            INSERT INTO {table_name}
            ({columns})
            VALUES
            ({values});
        """
    
    @classmethod
    def get_update_template(cls) -> str:
        return """
            UPDATE {table_name}
            set {dml}
            where {primary_key}={value}
        """
class Table:
    db_name:str
    name:str
    primary_key:Column
    columns:List[Column]
    meta:Dict[str,Column]

    @classmethod
    def read_from_sqltable(cls,sqltable:SQLTable, db_name:str) -> Table:
        table = Table()
        table.db_name = db_name
        table.name = sqltable.name
        columns:List[Column] = []
        meta = {}
        for sqlcolumn in sqltable.columns:
            debug(sqlcolumn)
            column = get_column_from_sql(sqlcolumn)
            meta[column.name] = column
        table.columns = meta.values()
        table.meta = meta
        for key in sqltable.primary_key:
            table.primary_key = get_column_from_sql(key)
            if key.name in meta.keys():
                meta[key.name].primary_key = True
                table.primary_key = meta[key.name]
        
        return table

    @classmethod
    def read_avro(cls, avro:dict, db_name:str,version:int) -> Table:
        table = Table()
        table.db_name = db_name
        table.name = avro['name']
        fields = avro['fields']
        meta = {}
        table.primary_key = None
        for field in fields:
            k = field['name']
            v = field['type']
            is_primary = field.get('primary_key',False)
            column = get_column_from_meta(k,v,is_primary)
            meta[k] = column
            if is_primary:
                table.primary_key = column
        if table.primary_key is None:
            raise Exception("no primary key in this table")
        table.columns = meta.values()
        return table



    def get_primary_date_column(self) -> Column:
        for column in self.columns:
            if column.nullable:
                continue
            elif isinstance(column,DateColumn):
                return column
            elif isinstance(column,DateTimeColumn):
                return column
        return None

    async def mock_insert(self,conn, hint) -> int: 
        tpl = DDLTemplate.get_insert_template() 
        debug(self.columns)
        columns = [column.name for column in self.columns if column.autoincrement is False]
        data = {column.name:fake(column, hint) for column in self.columns if column.autoincrement is False}
        sql = tpl.format(table_name=self.name,
                   columns=",".join(columns),
                   values=",".join(map(lambda value: ":{value}".format(value=value),columns)))

        debug(sql)
        data = jsonity(data)
        debug(data)
        r = await conn.execute(query=sql,values=data)
        return r

    async def mock_update(self, conn, hint, primary_id) ->  None:
        tpl = DDLTemplate.get_update_template()
        columns = [column.name for column in self.columns if column.autoincrement is False]
        data = {column.name:fake(column, hint) for column in self.columns if column.primary_key is False}
        items = []
        for k, v in data.items():
            column = self.meta[k]
            if isinstance(column,StringColumn) or \
                isinstance(column, DateColumn) or \
                    isinstance(column, DateTimeColumn):
                items.append("{k}='{v}'".format(k=k,v=v))
            else:
                items.append("{k}={v}".format(k=k, v=v))
        
        sql = tpl.format(table_name=self.name,
            dml = ",".join(items),
            primary_key = self.primary_key.name,
            value = primary_id
        )    
        debug(sql)
        r = await conn.execute(sql)
        return r        

    def get_ch_ddl(self) -> str:
        ch_columns = []
        for column in self.columns:
            line = "`{table_name}` {type}".format(table_name = column.name, type=column.to_ch_type())
            ch_columns.append(line)
        primary_date_column = self.get_primary_date_column()
        if primary_date_column is None:
            logger.error("{table_name} no primary date column".format(table_name=self.name))
            return None
        else:
            logger.success("{table_name} has primary date column".format(table_name=self.name))
            tpl = ClickhouseTableTemplate.get(CHTableEngine.ReplacingMergeTree)
            ddl = tpl.format(db_name=self.db_name,
                             table_name=self.name,
                             create_column=",".join(ch_columns),
                             primary_date = primary_date_column.name,
                             primary_key = self.primary_key.name)
            return ddl

    @classmethod
    def metadata(cls, database_url) -> dict:
        meta = sqlalchemy.MetaData()
        engine = sqlalchemy.create_engine(database_url)
        config = parse_url(database_url)
        meta.reflect(bind=engine)
        metatable = {}
        for sql_table_name, sql_table in meta.tables.items():
            table = Table.read_from_sqltable(sql_table,config['name'])
            key = "{db}/{table}".format(db=table.db_name,table=table.name)
            metatable[key] = table
        return metatable

    def normalize(self, data:dict) -> dict:
        r = {}
        for k, c in self.meta.items():
            v = data.get(k)
            
            if isinstance(c,DateColumn):
                if v is None:
                    r[k] = date.min
                elif isinstance(v,date):
                    r[k] = v
                else:
                    r[k] = date.fromisoformat(v)
            elif isinstance(c,DateTimeColumn):
                if v is None:
                    r[k] = datetime(2000,1,1)
                elif isinstance(v, datetime):
                    r[k] = v
                else:
                    r[k] = datetime.fromisoformat(v)
            elif isinstance(c, DecimalColumn):
                r[k] = str(v)
            elif isinstance(c, IntegerColumn):
                if c.nullable:  
                    r[k] = v
                else: 
                    if v is None: #当值为空且schema要求非空 设置默认值为0
                        r[k] = 0
                    else:
                        r[k] = v
            elif isinstance(c, StringColumn):
                r[k] = re.sub(r"^1\d{10}$",_fake.phone_number(),v)
            else:
                r[k] = v
        return r


    def get_default(self):
        default = {}
        for column in self.columns:
            if isinstance(column,DateTimeColumn):
                default[column.name] = datetime.now()
            else:
                default[column.name] = None
        return default
    




import configparser
config = configparser.ConfigParser()
config.read(".env")
import asyncio
from unimeta.pipeline import MysqlSource, ClickHouseSink, KafkaSink, Pipeline
from devtools import debug
from unimeta.pipeline import MetaServer
import uvloop

async def sync() -> None:
    source_url = config['source'].get('url')
    sink_url = config['sink'].get('url')
    meta_url = config['meta'].get('url')
    pipe = Pipeline(source_url, sink_url, meta_url)
    await pipe.start()
    pipe.sync_tables()
    await pipe.sync()

if __name__ == "__main__":
    uvloop.install()
    asyncio.run(sync())


from sqlalchemy import *
from databases import Database
import aiomysql
import sqlalchemy
from sqlalchemy.engine import reflection
import pytest
import functools
import asyncio
from devtools import debug
from unimeta.table import Table
from loguru import logger
import configparser
from unimeta.libs.liburl import parse_url
import pymysql
import time
import random

config = configparser.ConfigParser()
config.read(".env")
  
async def fake() -> None:
    database_url = config['mysql'].get("url")

    meta = Table.metadata(database_url)
    debug(meta)
    hint = {
        "email":"ascii_free_email",
        "phone_number":"phone_number",
        "first_name":"first_name",
        "last_name":"last_name"
    }
    async with Database(database_url) as database:
        while True:
            try:
                keys = list(meta.keys())
                key = random.choice(keys)
                debug(key)
                table = meta[key] 
                primary_id = await table.mock_insert(database, hint)
                await table.mock_update(database, hint, primary_id)
            except pymysql.err.DataError:
                logger.exception("what")

if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    res = loop.run_until_complete(fake())
    loop.close()


from pymysqlreplication import BinLogStreamReader
from unimeta.libs.liburl import parse_url
from pymysqlreplication.row_event import (
    DeleteRowsEvent,
    UpdateRowsEvent,
    WriteRowsEvent,
)
from typing import Dict
from devtools import debug
from unimeta.event import Event, EventType
from unimeta.table import Table
from clickhouse_driver import Client
import sys
import asyncio
import json
import random
from loguru import logger
import requests
from concurrent.futures import ProcessPoolExecutor
import aiohttp
from aiokafka import AIOKafkaProducer, AIOKafkaConsumer
from aioch import Client
import urllib


class MetaServer():
    name:str
    metaserver:str
    tables:dict

    def __init__(self,metaserver):
        self.meta = parse_url(metaserver)
        self.tables = {}
    
    async def reg(self,event):
        table_key = "{name}/{version}".format(name=event.name,version=event.version)
        if table_key in self.tables:
            return 
        port = self.meta['port']
        if port is None:
            port = 80
        meta_url = "http://{host}:{port}/api/meta/events/".format(
            host=self.meta['host'],
            port=self.meta['port'])
        data = {
            "name":event.name,
            "meta":event.avro(),
            "producer":self.meta['user'],
            'dataframe':event.table.name,
            'dataset': event.table.db_name,
            "scheme": self.meta['scheme'],
            "host": self.meta["host"],
            "port": self.meta["port"]
        }
        async with aiohttp.ClientSession() as session:
            async with session.post(meta_url, json=data) as response:
                self.tables[table_key] = event.table
                return event.table 

    async def get(self, event_name:str, version:int,dbname:str) -> Table:
        table_key = "{name}/{version}".format(name=event_name,version=version)
        if table_key in self.tables:
            return self.tables[table_key]
        name = urllib.parse.quote(event_name) 
        meta_url =   "http://{host}:{port}/api/meta/events/?name={name}&version={version}".format(
            host=self.meta['host'],
            port=self.meta['port'],
            name=name,
            version = version)
        debug(meta_url)    
        async with aiohttp.ClientSession() as session:
            async with session.get(meta_url) as resp:
                status_code = resp.status
                if status_code == 404:
                    logger.error("event {name} has no metatable".format(name=event_name))
                    return 
                data = await resp.text()
                
                data = json.loads(data)
                data = data['data']
                debug(data)
                event_id = data['id']
                meta = data['meta']
                table = Table.read_avro(meta,dbname,version)
                self.tables[table_key] = table
                return table


class Sink():
    
    async def publish(self):
        pass

class Source():

    async def subscribe(self):
        pass

    async def start(self):
        pass


class MysqlSource(Source):
    metatable:Dict[str,Table]
    stream: BinLogStreamReader

    def __init__(self,database_url, meta:MetaServer, server_id=None):
        if server_id is None:
            server_id = random.randint(1,1000)
        Source.__init__(self)
        settings = parse_url(database_url)
        settings['db'] = settings['name']
        del settings['name']
        del settings['scheme']
        self.metatable = Table.metadata(database_url)
        self.stream = BinLogStreamReader(connection_settings = settings, 
                                         server_id=server_id,
                                         blocking=True,
                                         only_events=[DeleteRowsEvent, WriteRowsEvent, UpdateRowsEvent])

    async def start(self):
        pass 

    async def subscribe(self):
        for binlogevent in self.stream:
            key = "{db}/{table}".format(db=binlogevent.schema, table=binlogevent.table)
            table = self.metatable.get(key)
            if table is None:
                continue
            for row in binlogevent.rows:
                if isinstance(binlogevent, DeleteRowsEvent):
                    event = Event.parse_binlog(table,EventType.DELETE,row)
                elif isinstance(binlogevent, UpdateRowsEvent):
                    event = Event.parse_binlog(table,EventType.UPDATE,row)
                elif isinstance(binlogevent, WriteRowsEvent):
                    event = Event.parse_binlog(table,EventType.INSERT,row)
                else:
                    raise Exception("event type not support")
                logger.info(event)
                if event is None:
                    continue
                yield event

    def close(self):
        self.stream.close()



class ClickHouseSink(Sink):
    
    def __init__(self, database_url):
        Source.__init__(self)
        self.client = Client.from_url(database_url)

    async def start(self):
        pass

    async def execute(self, query):
        await self.client.execute(query)
    
    async def publish(self, event):
        tpl = """ INSERT INTO {table_name}
            ({columns})
            VALUES
        """
        columns = [column.name for column in event.table.columns]
        sql = tpl.format(table_name="{db}.{table}".format(db=event.table.db_name, table=event.table.name),
                   columns=",".join(columns))
        try:
            await self.client.execute(sql,[event.data])
            logger.success("{sql} insert clickhouse".format(sql=sql))
        except:
            logger.error(event.data)
            logger.exception("what?")
        #logger.success("publish {event}　success".format(event=str(event)))


def delivery_report(err, msg):
    if err is not None:
        logger.error('Message delivery failed: {}'.format(err))
    else:
        logger.warning('Message delivered to {} [{}]'.format(msg.topic(), msg.partition()))


class KafkaSink(Sink):

    def __init__(self, database_url):
        Sink.__init__(self)
        settings = parse_url(database_url)
        loop = asyncio.get_event_loop()
        bootstrap = '{host}:{port}'.format(host=settings['host'],port=settings['port'])
        logger.info(bootstrap)
        self.producer = AIOKafkaProducer(
            loop=loop,
            bootstrap_servers = bootstrap
        )
        self.topic = settings['name']
        debug(self.producer)


    async def start(self):
        await self.producer.start()
   
    def execute(self, query):
        pass


    async def publish(self, event):
        debug(event)
        data = event.json()
        logger.info("write {topic}".format(topic=self.topic))
        await self.producer.send(self.topic, data.encode('utf-8'))

class KafkaSource(Source):

    def __init__(self, database_url, meta:MetaServer):
        Source.__init__(self)
        settings = parse_url(database_url)
        self.name = settings['name']
        user = settings['user']
        if user is None:
            user = "default"
        loop = asyncio.get_event_loop()
        bootstrap = '{host}:{port}'.format(host=settings['host'],port=settings['port'])
        self.consumer = AIOKafkaConsumer(
            self.name, loop=loop,
            bootstrap_servers = bootstrap,
            group_id = user
        )
        self.meta = meta

    async def start(self):
        await self.consumer.start()

    async def subscribe(self):
        async for msg in self.consumer:
            value = json.loads(msg.value.decode('utf8'))
            event_name = value.get("name")
            version = value.get("version")
            table_key = "{name}/{version}".format(name=event_name, version=version)
            table = await self.meta.get(event_name,version,self.name)
            if table:
                event = Event(event_type=value.get('type'),
                    name=value.get('name'),
                    data=value.get('data'),
                    id=value.get('id'),
                    table = table,
                    version = value.get('version'),
                    ctx = value.get('ctx'))
                debug(event)
                yield event
            else:
                logger.error("{key} has no metatable".format(key=table_key))





class Pipeline():
    sink:Sink
    source:Source
    meta:dict
    metaserver:str
    
    def __init__(self,source:str, sink:str, meta:str):
        mconf = parse_url(meta)
        if mconf['scheme'] == 'unimetad':
            self.metaserver = MetaServer(meta)
        else:
            raise Exception("unregister meta")

        sconf = parse_url(source)
        if sconf['scheme'] == 'mysql':
            self.source = MysqlSource(source, meta=self.metaserver)
        elif sconf['scheme'] == 'kafka':
            self.source = KafkaSource(source, meta=self.metaserver)
        else:
            raise Exception("unregister source")
        
        dconf = parse_url(sink)
        if dconf['scheme'] == 'clickhouse':
            self.sink = ClickHouseSink(sink)
        elif dconf['scheme'] == 'kafka':
            self.sink = KafkaSink(sink)
        else:
            raise Exception("unregister sink")
        


    async def sync_tables(self):
        if isinstance(self.source, KafkaSource):
            return
        tables = self.source.metatable.values()
        for table in tables:
            ddl = table.get_ch_ddl()
            if ddl is not None:
                await self.sink.execute(ddl)

    def rebuild_table(self, table):
        source_tables = self.source.metatable.values()
        for item in source_tables:
            if table == item.name:
                delete_stmt = "drop table {db}.{table}".format(db=item.db_name,table=table)
                create_stmt = item.get_ch_ddl()
                self.sink.execute(delete_stmt)
                self.sink.execute(create_stmt)

    async def start(self):
        await self.source.start()
        await self.sink.start()

    async def sync(self):
        async for event in self.source.subscribe():
            debug(event)
            await self.metaserver.reg(event)
            await self.sink.publish(event)